#include <stdlib.h>
#include <string.h>

#include "../common/spi.h"
#include "../common/flash.h"

/* Instruction Set */

#define READ_DATA					0x03
#define FAST_READ					0x0b
#define SECTOR_ERASE				0xd8
#define BULK_ERASE					0xc7
#define PAGE_PROGRAM				0x02
#define READ_STATUS_REGISTER		0x05
#define WRITE_STATUS_REGISTER		0x01
#define WRITE_ENABLE				0x06
#define WRITE_DISABLE				0x04
#define JEDEC_ID					0x9f
#define READ_SIGNATURE				0xab

/* STATUS bits */
#define STATUS_BUSY					0x01
#define STATUS_WEL					0x02


#define NUM_SECTORS 512			//  Changed sectors from 32 to 512 for Spansion -DJH
#define SECTOR_SIZE 0x1000 		//  Changes sector size from 0x10000 to 0x1000 - DJH


static int s25fl116k_write_enable(const struct flash_info *fi);
static int s25fl116k_write_status(const struct flash_info *fi, uint8_t status, uint8_t dummy);

static int s25fl116k_open(struct flash_info *fi)
{
	fi->size = 0x200000;
	fi->number_of_regions = 1;
	fi->erase_block_regions = malloc(fi->number_of_regions * sizeof (struct erase_block_region));

	if (fi->erase_block_regions == NULL)
		return -1;

	fi->erase_block_regions[0].block_size = SECTOR_SIZE;
	fi->erase_block_regions[0].number_of_blocks = NUM_SECTORS;

	/* Clear all block protection bits */
	s25fl116k_write_enable(fi);
	s25fl116k_write_status(fi, 0, 0);

	return 0;
}

static int s25fl116k_close(struct flash_info *fi)
{
	free(fi->erase_block_regions);
	return 0;
}

static int s25fl116k_write_enable(const struct flash_info *fi)
{
	return generic_write_enable(fi, WRITE_ENABLE);
}

static int s25fl116k_write_disable(const struct flash_info *fi)
{
	return generic_write_disable(fi, WRITE_DISABLE);
}

static int s25fl116k_read_mid_did(struct flash_info *fi, uint8_t *mid, uint8_t *did)
{
	/* s25fl116k uses JEDEC_ID.  */
	return -1;
}

static int s25fl116k_read_uid(const struct flash_info *fi, uint64_t *uid)
{
	return -1;
}

static int s25fl116k_read_jedec_id(const struct flash_info *fi, uint8_t *mid,
		uint8_t *memory_type_id, uint8_t *capacity_id)
{
	return generic_read_jedec_id(fi, JEDEC_ID, mid, memory_type_id, capacity_id);
}

static int s25fl116k_read_status(const struct flash_info *fi, uint8_t *status)
{
	uint8_t tbuf[1];
	uint8_t rbuf[1];
	int count;

	select_flash();

	count = 0;
	assign_instruction(fi, tbuf, READ_STATUS_REGISTER, &count);
	spi_send(tbuf, count);

	spi_recv(rbuf, 1);
	*status = rbuf[0];

	unselect_flash();

	return 0;
}

static int s25fl116k_not_busy(const uint8_t status)
{
	return (status & STATUS_BUSY) ? 0 : 1;
}

static int s25fl116k_wait_ready(const struct flash_info *fi)
{
	uint8_t tbuf[1];
	int count;

	select_flash();

	count = 0;
	assign_instruction(fi, tbuf, READ_STATUS_REGISTER, &count);
	spi_send(tbuf, count);

	spi_recv_until(s25fl116k_not_busy);

	unselect_flash();

	return 0;
}

static int s25fl116k_write_status(const struct flash_info *fi, uint8_t status, uint8_t dummy)
{
	uint8_t tbuf[2];
	int count;
	uint8_t old_status;

	s25fl116k_write_enable(fi);

	s25fl116k_read_status(fi, &old_status);
	if (!(old_status & STATUS_WEL))
		return -1;

	select_flash();

	count = 0;
	assign_instruction(fi, tbuf, WRITE_STATUS_REGISTER, &count);
	tbuf[count++] = status;
	spi_send(tbuf, count);

	unselect_flash();

	s25fl116k_wait_ready(fi);

	return 0;
}

static int s25fl116k_read(const struct flash_info *fi, uint32_t addr, uint8_t *buf, int size)
{
	uint8_t read_insn;
	uint8_t tbuf[5];
	int count;

	/*  TODO frequency will limit instruction chosen.  */
	read_insn = READ_DATA;

	select_flash();

	count = 0;
	assign_instruction(fi, tbuf, read_insn, &count);
	assign_address(fi, tbuf, addr, &count);

	if (read_insn == FAST_READ)
		count++;

	spi_send(tbuf, count);

	spi_recv(buf, size);

	unselect_flash();

	return 0;
}

/* TODO make erase functions generic */

/* Internal function used by s25fl116k_erase and s25fl116k_erase_chip */

static int s25fl116k_erase_internal(const struct flash_info *fi,
		uint32_t addr, erase_type_t erase_type)
{
	uint8_t tbuf[4];
	int count;
	uint8_t status;
	uint8_t insn;
	uint32_t addr_mask;

	switch (erase_type)
	{
	case ERASE_4KB:
		insn = SECTOR_ERASE;
		addr_mask = 0xfffff000;
		break;

	case ERASE_CHIP:
		insn = BULK_ERASE;
		addr_mask = 0;
		break;

	default:
		/* bad erase type */
		return -1;
	}

	s25fl116k_write_enable(fi);

	s25fl116k_read_status(fi, &status);
	if (!(status & STATUS_WEL))
		return -1;

	select_flash();

	count = 0;
	assign_instruction(fi, tbuf, insn, &count);
	if (erase_type != ERASE_CHIP)
		assign_address(fi, tbuf, addr & addr_mask, &count);
	spi_send(tbuf, count);

	unselect_flash();

	s25fl116k_wait_ready(fi);

	return 0;
}

static int s25fl116k_erase(const struct flash_info *fi, uint32_t addr, int size)
{
#define ERASE_TYPE ERASE_4KB
#define ERASE_SIZE (4 * 1024)

	size += addr % ERASE_SIZE;
	addr -= addr % ERASE_SIZE;

	while (size > 0)
	{
		s25fl116k_erase_internal(fi, addr, ERASE_TYPE);
		addr += ERASE_SIZE;
		size -= ERASE_SIZE;
	}

	return 0;
}

static int s25fl116k_erase_chip(const struct flash_info *fi)
{
	return s25fl116k_erase_internal(fi, 0, ERASE_CHIP);
}

#define PAGE_SIZE 256

static int s25fl116k_page_program (struct flash_info *fi,
		uint32_t addr, const uint8_t *buf, int size)
{
	uint8_t tbuf[4 + size];
	int count;
	uint8_t old_status;

	if ((addr % PAGE_SIZE) + size > PAGE_SIZE)
		return -1;

	s25fl116k_write_enable (fi);

	s25fl116k_read_status (fi, &old_status);
	if (!(old_status & STATUS_WEL))
		return -1;

	select_flash ();

	count = 0;
	assign_instruction (fi, tbuf, PAGE_PROGRAM, &count);
	assign_address (fi, tbuf, addr, &count);
	memcpy (tbuf + count, buf, size);
	spi_send (tbuf, count + size);

	unselect_flash ();

	s25fl116k_wait_ready (fi);

	return 0;
}

/* Program arbitary size */
static int s25fl116k_program (struct flash_info *fi,
		uint32_t addr, const uint8_t *buf, int size)
{
	int program_size;

	while (size > 0)
	{
		if (addr % PAGE_SIZE != 0)
			program_size = ((PAGE_SIZE - addr % PAGE_SIZE) < size
					? (PAGE_SIZE - addr % PAGE_SIZE) : size);
		else if (size >= PAGE_SIZE)
			program_size = PAGE_SIZE;
		else
			program_size = size;

		s25fl116k_page_program (fi, addr, buf, program_size);

		addr += program_size;
		buf += program_size;
		size -= program_size;
	}

	return 0;
}

static int s25fl116k_reset(const struct flash_info *fi)
{
	/* TODO implement it */
	return -1;
}

struct flash_info s25fl116k_info =
{
	"s25fl116k",	/* name */
	"Spansion",	/* name */
	0x01,/* manufacturer ID m25 = 0x20 */								// Spansion manID = 0x01 - DJH
	/* device ID.  s25fl116k's DID is 16-bit and is "memory type ID, capacity Id".
	 * so we just check them instead of device ID.  */
	0xff,      /*0xff*/
	0x40,		/* memory type ID m25 = 0x20 */ 						//Spansion memID = 0x40 - DJH
	0x15,		/* capacity ID m25 = 0x15 */							//Spansion capID = 0x15 - DJH
	STANDARD,	/* supported modes */

	/* The following three fields are better be
	   initialized in flash_open functions.  */
	0,			/* size in Byte */
	0,			/* number_of_regions */
	NULL,		/* erase_block_regions */

	25,		/* max_freq */
	0,		/* max_quad_freq */
	20,		/* max_read_data_freq */

	0,		/* cpha */
	0,		/* cpol */
	0,		/* lsb_first */
	0,		/* start_on_mosi */

	0,		/* frequency */
	STANDARD,	/* current_mode */

	s25fl116k_open,
	s25fl116k_close,

	s25fl116k_read_mid_did,
	s25fl116k_read_uid,
	s25fl116k_read_jedec_id,
	s25fl116k_read_status,
	NULL, /* read_status2 */
	s25fl116k_write_status,
	NULL, /* flash_enable_quad_mode */
	NULL, /* flash_disable_quad_mode */
	s25fl116k_read,
	s25fl116k_write_enable,
	s25fl116k_write_disable,
	s25fl116k_erase,
	s25fl116k_erase_chip,
	s25fl116k_program,
	s25fl116k_reset,
};
